:- [altp].
:- dynamic belief/2.
:- dynamic block/1.
:- op(950, xfy, and).
:- op(700, xfx, in).

% Commands to use in Tetris prompt
start :- open('tetris-out.txt', write, _, [alias(output)]),
         thread_create(run, _, [alias(reasoner)]),
	 do(start).
quit  :- do(quit).
drop  :- do(drop).
do(X) :- write(':'), write(X), nl.

% agent actions
action(left).
action(right).
action(down).
action(rotate).		% Rot+
action(\rotate).	% Rot-
action(drop).

make_obs(blocks(Blocks)) :- !, make_obs(locked), send(Blocks).
make_obs(X) :- recorded(obs, locked) -> recordz(obs1, X) ; recordz(obs, X).

observations(_, Obs) :- get_recorded(obs, Obs, []).

% get a list of recorded values
% and then erase them
get_recorded(Key, [X|Xs], Refs) :-
  recorded(Key, X, Ref), \+ member(Ref, Refs), !,
  get_recorded(Key, Xs, [Ref|Refs]).
get_recorded(Key, [], [Ref|Refs]) :-
  erase(Ref), get_recorded(Key, [], Refs).
get_recorded(_, [], []).

% send message to reasoner thread
send(X) :- thread_send_message(reasoner, X).

% run thread
run :- repeat, observations(_, [next(X),locked]), !, run(X).

% thread main loop
run(X) :- thread_get_message(Blocks), run(X, Blocks), !,
          (belief(_, next(Y)) ; Y=nil), reset, !, run(Y).

run(X, B) :- catch((
  retractall(block(_)), assert_blocks(B),
  (is_stream(output) -> true ;
    current_output(Out), set_stream(Out,alias(output))),
  with_output_to(output, print_blocks(B)),
  (X=nil ; assert_belief(0, piece(X))),
  forall(axiom(Form), assert_belief(0, Form)),
  forall(recorded(obs1, Form, Ref), (recordz(obs, Form), erase(Ref))),
  repeat,
    with_output_to(output, step(I)),
    (belief(I, locked) ; once(run1(I)), fail)
), E, (print_message(error, E), fail)).

run1(I) :-
  belief(I, do(A)) -> do(A) ;
  belief(I, goal(G)) -> planner_goal(I, G) ;
  belief(I, plan(S0, G, Steps)),
  belief(I, planned(S0, G)),
  new_fact(belief(I, planned_next(S0, G))), !,
  transition(S0, down, S1),
  next_plan(S1, G, Steps, Steps1),
  (belief(I, failed(S0, G)) -> delete_facts(belief(I, plan(S0, G, _))) ; true),
  Steps1 \= [], Steps1 \= [drop],
  assert_belief(I, plan(S1, G, Steps1)) ;
  \+belief(I, plan(_,_,_)),
  choose_goal(I).

choose_goal(I) :-
  (belief(I, planned(_, G)), G=locked(_) ;
   must_hurry_goal(I, G)  ;
   once(I>10 ; belief(I, goal_ready)), best_goal(I, G)),
  \+belief(I, not(goal(G))),
  assert_belief(I, goal(G)).

planner_goal(I, G) :-
  belief(I, consistent), S0=at(_,_), belief(I, S0),
  new_fact(belief(I, planned(S0, G))),
  (make_plan(S0, G, Steps) ->
    print_plan(G, Steps), assert_belief(I, plan(S0, G, Steps)) ;
    assert_belief(I, not(goal(G)))),
  delete_facts(belief(I, goal(G))).

next_plan(S1, G, Steps, Steps1) :-
  select(down, Steps, Steps1), verify_plan(S1, G, Steps1).
next_plan(S1, _, Steps, Steps) :-
  append(Steps1, [drop], Steps), verify_plan(S1, _, Steps1).

print_plan(G, Steps) :-
  write('Plan g='), print(G), nl,
  forall(member(X, Steps),
    ((X=goal(_) -> write('...') ; write(X), write(' ')))),
  nl, nl.

assert_blocks([P|Ps]) :- assert(block(P)), assert_blocks(Ps).
assert_blocks([]).

assert_belief(I, Form) :- new_fact(belief(I, Form)), ! ; true.

% define belief structure
time_step(belief(I,_),I).

% -------------------
%   Inference rules
% -------------------

rule([],  belief(I, now(I))).
rule([obs(I,A)],  belief(I,A)).

% Extended modus ponens
rule([belief(I, Ps->Qs), all(conjunct(P,Ps), beliefx(I,P)), conjunct(Q,Qs)],
     belief(I+1, Q)).

% goal generation rules
rule([\+belief(I-1,goal_ready)|Prems], Concl) :-
  Concl=belief(I,_), goalgen_rule(Prems, Concl).

% Contradictions
rule([belief(I, not(A)), belief(I, A)],  belief(I+1, contra(I, A, not(A)))).
rule([belief(I, A), belief(I, B), A\=B], belief(I+1, contra(I, A, B))) :-
  (A=at(_,_) ; A=piece(_)), copy_term(A, B).

% Inheritance
rule([belief(I, A), A\=now(_), A\=delete(_), A\=do(_), A\=consistent,
      \+belief(I, contra(I-1, A, _)),
      \+belief(I, contra(I-1, _, A)),
      \+belief(I, delete(A)),
      \+(A=at(_,_) and belief(I, do(_)))],
     belief(I+1, A)).

% mark consistent
rule([\+inconsistent(I)], belief(I, consistent)).

% insert transitions
rule([belief(I, S1), transition(S1,A,S2)],
     belief(I, transition(S1,A,S2))) :- S1=at(_,_).
rule([belief(I, plan(S0,_,[A|_])), transition(S0,A,S1)],
     belief(I, transition(S0,A,S1))).

% plan execution rules
rule([belief(I,consistent)|Prems], Concl) :-
  Concl=belief(I,_), planexec_rule(Prems, Concl).

beliefx(_, A<B) :- integer(A), integer(B), A<B.
beliefx(_, A\=B) :- A\=B.
beliefx(I, all(Cond, P)) :- all(belief(I,Cond), belief(I,P)).
beliefx(I, A) :- query(belief(I, A)).

inconsistent(I) :- S1=at(_,_), S2=at(_,_), belief(I,S1), belief(I,S2), S1\=S2.

% Axioms
axiom(height(X,H)) :- height(X,H).
axiom(piece(X) -> fixed(B)) :- B=tetromino(X,_), fixed(B).
axiom(at(tetromino(X,_),_) -> piece(X)).
axiom(Form) :- goalgen_axiom(Form).
axiom(Form) :- planexec_axiom(Form).

% discard unneeded transitions
axiom(at(_, pos(_,Y)) and T -> delete(T)) :- T=transition(_,_,at(_,pos(_,Y+1))).

axiom(piece(X) -> deadline(19-N+4)) :-
  above_surface(tetromino(X,1), pos(4,19), N).

above_surface(B,P,0)   :- transition(at(B,P), down, locked(_)).
above_surface(B,P,N+1) :- transition(at(B,P), down, at(B,Q)), above_surface(B,Q,N).

% Basic Tetris logic
% see 3.3 in the report

% symbol, number of orientations
tetromino(i,2).
tetromino(j,4).
tetromino(l,4).
tetromino(o,1).
tetromino(s,2).
tetromino(t,4).
tetromino(z,2).

fixed(tetromino(X,I)) :- tetromino(X,N), between(1,N,I).

rotate(tetromino(X,I),tetromino(X,J)) :- tetromino(X,N), between(2,N,J), succ(I,J).
rotate(tetromino(X,N),tetromino(X,1)) :- tetromino(X,N).

% defines occupied area of each fixed tetromino
occupies(tetromino(i,1),X,1) :- between(0,3,X).
occupies(tetromino(i,2),1,Y) :- between(0,3,Y).
occupies(tetromino(o,1),X,Y) :- between(1,2,X), between(0,1,Y).
occupies(tetromino(j,1),2,0).  occupies(tetromino(j,3),0,2).
occupies(tetromino(l,1),0,0).  occupies(tetromino(l,3),2,2).
occupies(tetromino(t,1),1,0).  occupies(tetromino(t,3),1,2).
occupies(tetromino(s,1),0,0).  occupies(tetromino(s,1),1,0).
occupies(tetromino(s,1),1,1).  occupies(tetromino(s,1),2,1).
occupies(tetromino(z,1),X,Y) :- occupies(tetromino(s,1),Z,Y), reflect(X,Z).
occupies(tetromino(W,2),X,Y) :- flip(V,W), occupies(tetromino(V,1),Y,X).
occupies(tetromino(W,4),X,Y) :- flip(V,W), occupies(tetromino(V,3),Y,X).
occupies(tetromino(W,1),X,1) :- occupies(tetromino(W,3),X,1).
occupies(tetromino(W,3),X,1) :- tetromino(W,4), between(0,2,X).

reflect(0,2).
reflect(1,1).
reflect(2,0).

flip(j,l).
flip(l,j).
flip(s,z).
flip(z,s).
flip(t,t).

% move between cells
move(pos(X1,Y1),M,N,pos(X2,Y2)) :- plus(X1,M,X2), plus(Y1,N,Y2).

blocks(B,P,Q) :- occupies(B,M,N), move(P,M,N,Q).

X in union(A,B)  :- X in A ; X in B.
Q in blocks(B,P) :- blocks(B,P,Q).
P in blocks      :- block(P).
pos(X,0) in row0 :- between(1,10,X).

height(X,B,H) :- pos(X,H) in union(B,row0), 0=<H, H=<22,
                 \+ (pos(X,Y) in B, Y>H).

height(X,H) :- height(X,blocks,H).

possible(at(B,P)) :-
  P=pos(_,Y), Y<20,
  \+ (pos(X,Y1) in blocks(B,P), (X<1 ; X>10 ; Y1<1 ; block(pos(X,Y1)))).

% PLANNING
% see 3.4 in the report

verify_plan(S, S, []).
verify_plan(S1, G, [A|Plan]) :-
  transition(S1, A, S2), verify_plan(S2, G, Plan).

% Forward best-first search
plan(S, S, _, _, []).
plan(S1, G, HG, Path, [A|Plan]) :-
  best_first(S1, A, S2, HG), \+member(S2, Path),
  plan(S2, G, HG, [S1|Path], Plan).

best_first(S1, A, S2, G) :-
  Gen=(transition(S1, A, S2), heuristic(S2, G, N)),
  setof(N, A^S2^Gen, Ns), member(N, Ns), Gen.

% only follow heuristic optimal path
plan2(S0, G, Plan) :-
  heuristic(S0, G, N), length(Plan, N), plan2(S0, G, N, Plan).
plan2(S, S, 0, []).
plan2(S1, G, N, [A|Plan]) :-
  transition(S1, A, S2), heuristic(S2, G, M), succ(M,N),
  plan2(S2, G, M, Plan).

% Transition function
transition(S1,down,locked(S1)) :- move(S1,down,S2), \+possible(S2), possible(S1).
transition(S1,A,S2) :- var(S1), !, move(S1,A,S2), possible(S1).
transition(S1,A,S2) :- move(S1,A,S2), possible(S2).

move(at(B,P), rotate,  at(C,P)) :- rotate(B,C), B\=tetromino(o,1).
move(at(B,P), \rotate, at(C,P)) :- rotate(C,B), B\=tetromino(o,1).
move(at(B,P), right, at(B,Q)) :- move(P,1,0,Q), fixed(B).
move(at(B,P), left,  at(B,Q)) :- move(P,-1,0,Q), fixed(B).
move(at(B,P), down,  at(B,Q)) :- move(P,0,-1,Q), fixed(B).

% heuristic distance to goal
heuristic(G, G, 0).
heuristic(at(B,pos(X1,Y1)), at(C,pos(X2,Y2)), N) :-
  Y1 >= Y2, rotations(B,C,M), N is abs(X1-X2) + Y1-Y2 + M.
heuristic(S1, locked(S2), N) :- heuristic(S1, S2, M), succ(M,N).

rotations(X,X,0).
rotations(X,Y,1) :- rotate(X,Y) ; rotate(Y,X).
rotations(X,Z,2) :- rotate(X,Y), rotate(Y,Z).

% -------------------
%   Goal generation
% -------------------

% bottom edge of the current piece
goalgen_rule(
  [belief(I, fixed(B)), occupies(B,M,N), \+occupies(B,M,N-1)],
  belief(I+1, bottom_edge(B,M,N))).

% the piece fits after line clears
goalgen_rule(
  [belief(I, lines(B,pos(X,Y),Ys)),
   belief(I, bottom_edge(B,M,N)), belief(I, height(X+M,Y+N-1)),
   all(belief(I, bottom_edge(B,M1,N1)) and M1\=M and
       \+all(occupies(B,M1,N2), member(Y+N2, Ys)),
       belief(I, height(X+M1, Y+N1-1)))],
  belief(I+1, fits(B,pos(X,Y)))).

% goal ready when fits(b,p) doesn't increase height
goalgen_rule(
  [belief(I, fits(B,P)), belief(I, height(_,H)),
   \+ (belief(I, height(_,H1)) and H1>H),
   all(pos(_,Y) in blocks(B,P), Y=<H)],
  belief(I+1, goal_ready)).

goalgen_axiom( piece(X) -> lines(B,P,Ys) ) :-
  B=tetromino(X,_), bagof(Y, line(B,P,Y), Ys).

% positions where the piece fits on top of the surface
goalgen_axiom(
  fixed(B) and all(bottom_edge(B,M,N), height(X+M, Y+N-1)) and Y<20
  -> fits(B,pos(X,Y)) ).

best_goal(I, locked(at(B,P))) :-
  belief(I, piece(X)), B=tetromino(X,_),
  between(1,16,Y),
  belief(I, fits(B,P)),
  \+ (pos(_,Y1) in blocks(B,P), Y1>Y).

must_hurry_goal(I, locked(at(B,P))) :-
  belief(I, at(B1, pos(_,Y1))),
  belief(I, deadline(Y2)), Y1-1 =< Y2, !,
  B1=tetromino(X1,_), B=tetromino(X1,_),
  P=pos(X,Y),
  between(0,19,Y),
  (belief(I, fits(B,P)) ;
   belief(I, bottom_edge(B,M,N)),
   query(belief(I, height(X+M, Y+N-1))),
   possible(at(B,P))).

% tetromino completes a line
line(T,P,Y) :- B=blocks(T,P), fillsrow(B,Y),
               forall(Q in B, (visible(Q), \+block(Q))).
visible(pos(X,Y)) :- 0<X,X<11, 0<Y,Y=<20.

% tetromino blocks fills row y
fillsrow(B,Y) :-
  filledto(Y,X1), filledfrom(Y,X4), succ(X1,X2), succ(X3,X4),
  pos(X2,Y) in B, pos(X3,Y) in B,
  forall(pos(X,Y) in B, (X2=<X, X=<X3)).
fillsrow(B,Y) :-
  filledto(Y,X1), X1>=6, succ(X1,X2), pos(X2,Y) in B, pos(10,Y) in B,
  forall(pos(X,Y) in B, (X2=<X, X=<10)).
fillsrow(B,Y) :-
  filledfrom(Y,X2), X2=<5, succ(X1,X2), pos(1,Y) in B, pos(X1,Y) in B,
  forall(pos(X,Y) in B, (1=<X, X=<X1)).

% filled from left wall to x
filledto(Y,X) :- block(pos(1,Y)), filledto(Y,X,1).
filledto(Y,X,X1) :- succ(X1,X2), block(pos(X2,Y)), !, filledto(Y,X,X2).
filledto(_,X,X).

% filled from x to right wall
filledfrom(Y,X) :- block(pos(10,Y)), filledfrom(Y,X,10).
filledfrom(Y,X,X2) :- succ(X1,X2), block(pos(X1,Y)), !, filledfrom(Y,X,X1).
filledfrom(_,X,X).

% PLANNER
% Make a plan
make_plan(S0, locked(G), Steps) :-
  make_plan1(S0, G, Steps1), !, (downs_to_drop(Steps1,Steps) ; Steps=Steps1).
make_plan(S0, G, Steps) :-
  make_plan1(S0, G, Steps).

make_plan1(S0, G, Plan) :- plan1(S0, G, Plan).
make_plan1(S0, G, [A|[goal(SubG)|Plan]]) :- 
  subgoal(SubG, G), SubG\=G, heuristic(S0, SubG, N), N>1,
  (plan(S0, _, SubG, [], [A]) -> plan1(SubG, G, Plan)),
  \+repeat_list(down, Plan).

% fails quickly
plan1(S0, G, Plan) :- heuristic(S0, G, N), N<8, !, plan2(S0, G, Plan).
plan1(S0, G, Plan) :-
  length(Plan1, 5), plan(S0, SN, G, [], Plan1), !,
  append(Plan1, Plan2, Plan),
  plan1(SN, G, Plan2).

subgoal(G, G) :-
  G=at(B,P), \+ (pos(X,Y) in blocks(B,P), height(X,H), H+2 > Y), !.
subgoal(SubG, S2) :-
  transition(S1, down, S2), !, subgoal(SubG, S1).
subgoal(SubG, G) :-
  between(1,3,N), length(Shifts,N), repeat_list(A, Shifts),
  member(A, [right, left]),
  member(Plan, [Shifts, [\rotate|Shifts], [rotate|Shifts]]),
  verify_plan(G, S, Plan), 
  transition(SubG, down, S), subgoal(_, SubG).

downs_to_drop([down|[down|Xs]], [drop]) :- repeat_list(down, Xs), !.
downs_to_drop([X|Xs], [X|Ys]) :- downs_to_drop(Xs, Ys).
downs_to_drop([], []).

repeat_list(_, []).
repeat_list(X, [X|Xs]) :- repeat_list(X, Xs).

% ------------------
%   Plan execution
% ------------------

% Execute planned action
planexec_rule(	[belief(I, plan(Pre,_,[Act|_])), action(Act), Act\=down,
		 belief(I, Pre)],
		belief(I, do(Act))).
planexec_rule(	[belief(I, plan(Pre,_,[down|[down|_]])),
		 belief(I, Pre)],
		belief(I, do(down))).
% next step is a goal
planexec_rule(	[belief(I, plan(Pre,_,[goal(G)|_])),
		 \+belief(I, plan(_,G,_)),
		 belief(I, Pre)],
		belief(I, goal(G))).

planexec_axiom( do(A) and transition(S1,A,S2) and S1 -> S2 ).
planexec_axiom( do(drop) and plan(_,G,[drop]) -> G and not(goal(_)) ).

% resolve state contradictions
planexec_axiom(
  now(I) and contra(I-1, S1,S2) and transition(S1,down,S2) -> S2).
planexec_axiom(
  now(I) and contra(I-1, S2,S3) and transition(S1,A,S2)
                                and transition(S3,A,S4) -> S4) :-
  S1=at(B, pos(X,Y)), S2=at(_, pos(_,Y)),
  S3=at(B, pos(X,Y-1)).

% update plan
% Plan0 and transition(S0,A)=S1 and ((S0 and do(A)) or S1) -> ...
planexec_axiom(
  plan(S0, G, [A|Steps]) and transition(S0,A,S1) and Ps
  -> plan(S1, G, Steps) and delete(plan(S0, G, _)) ) :-
  Ps=(S0 and do(A)) ; Ps=(S1 and consistent).
planexec_axiom(
   plan(Pre, G, [goal(SubG)|Steps]) and plan(_, SubG, _)
   -> plan(SubG, G, Steps) and delete(plan(Pre, G, _)) ).

planexec_axiom(
  plan(S0, G, [down|[down|Steps]])
  -> plan(S2, G, Steps) and delete(S0) and delete(plan(S0, G, _))
                        and delete(S1) and delete(plan(S1, G, _)) ) :-
  S0=at(B, pos(X,Y)),
  S1=at(B, pos(X,Y-1)),
  S2=at(B, pos(X,Y-2)).

% delete plan when goal is reached
planexec_axiom( plan(_, G, _) and G and consistent -> delete(plan(_, G, _)) ).

% Plan failures
% Tetris unreachable state
planexec_axiom(
  plan(at(B,pos(X,Y)), G, _) and at(_,pos(_,Y1)) and consistent and
  Y1 < Y -> failed(at(B,pos(X,Y)), G) ).
% subgoal failed
planexec_axiom(
  plan(Pre, G, [goal(SubG)|_]) and not(goal(SubG))
  -> failed(Pre, G) and not(goal(G)) ).

planexec_axiom(
  plan(Pre, G, [A|_]) and A\=down and failed(Pre, G)
  -> delete(plan(_,_,_)) ).

% Pretty printing

print_blocks(Ps) :- write('Blocks:'), nl, print_blocks(Ps,20).
print_blocks(Ps,Y2) :- print_row(Ps,Y2), succ(Y1,Y2), print_blocks(Ps,Y1).
print_blocks(_,0).
print_row(Ps,Y) :- member(pos(_,Y),Ps), !, print_row(Ps,Y,1), nl.
print_row(_,_).
print_row(_,_,11).
print_row(Ps,Y,X1) :- X1<11,
  (member(pos(X1,Y), Ps) -> formatcell(X1,Y) ; write('  ~  ')),
  succ(X1,X2), print_row(Ps,Y,X2).
formatcell(X,Y) :- write('|'), write(X), write(','), write(Y), (Y<10 -> write(' ') ; true).

portray(tetromino(X,I)) :- atom(X), upcase_atom(X,Y), write(Y), write('_'), print(I).
portray(height(X,H)) :- print(height(X)=H).
portray(all(Cond, P)) :- write('(ALL '), print_formula(Cond),
                         write(': '),    print_formula(P), write(')').
portray(pos(X,Y)) :- write('('), print(X), write(','), print(Y), write(')').
