% Active logic theorem prover
% 2009-2010 VN

% time_step(?Belief, ?I)
%   Return/unify time in belief. Example: time_step(belief(I,_), I).
%
% Rules           rule(Premises, Concl)
% Observations    obs(I, Form)

:- load_foreign_library(bucketarray).
:- bucketarray(1000).     % hash table size

% Number of steps to keep in memory after a completed step.
history_size(1).

step(I) :- next_step(I), record_obs(I), !, forall(apply_rule(I,_), true),
           nl, time_step(Belief, I),
           forall(call(Belief), (print_numvars(Belief),nl)),
           %nl, statistics,
           end_step(I).

step(I,N) :- I<N, step(I), succ(I,J), step(J,N), !.
step(I,I) :- step(I).

now(I) :- flag(now,I,I).

next_step(0) :- now(0), flag(step0,0,1).
next_step(I) :- now(Now), succ(Now,I), flag(now,Now,I).

end_step(I) :- history_size(Size), !, J is I-Size, forget_step(J).

% Free all the history, go back to step 0.
reset :- now(I), forget_step(I), succ(J,I), !, flag(now,I,J), reset.
reset :- now(0), flag(step0,_,0).

% Free beliefs and observations of step I.
forget_step(I) :- time_step(Belief,I), delete_facts(Belief), erase_obs(I).

% Observations in recorded database.
% If observations/2 is defined, it should return a list of observed formulas
% in Obs that are recorded under key I.
record_obs(_) :- \+ current_predicate(observations/2), !.
record_obs(I) :- observations(I,Obs), record_obs(I,Obs).
record_obs(_,[]).
record_obs(I,[Form|Obs]) :- recordz(I,Form), record_obs(I,Obs).
erase_obs(I) :- forall(recorded(I,_,Ref), erase(Ref)).

% Infer one new belief at time step I.
% Backtracking over all solutions will generate the belief set of time I.
% apply_rule(+I, -Belief)
apply_rule(I,Belief) :-
  rule(Prems,Concl), time_step(Concl,J), eq(I,J),
  premises(Prems), eval(Concl,Belief), new_fact(Belief).
premises([]).
premises([P1|Prems]) :- eval(P1,P2), query(P2), premises(Prems).

% could be extended with disjunctions and axioms
query(not(Goal)) :- !, \+ query(Goal).
query(\+ Goal)   :- !, \+ query(Goal).
query(obs(I,Form)) :- \+ current_predicate(obs/2), !, recorded(I,Form).
query(Belief) :- time_step(Belief, I), time_step(Pattern, I),
  pattern(Belief, Pattern), !, call(Pattern), equals(Belief, Pattern).
query(Goal) :- call(Goal).

and(P,Q) :- query(P), query(Q).

% Pattern is a generic term which can be unified with Term;
% integers and operations in Term are variables in Pattern.
% pattern(+Term, ?Pattern)
pattern(T,P) :- compound(T), functor(T,F,N), \+ operation(T), !,
                functor(P,F,N), pattern(T,P,1).
pattern(_,P) :- var(P), !.
pattern(T,T).
pattern(T,P,I) :- arg(I,T,A1), !, arg(I,P,A2), pattern(A1,A2),
                  succ(I,J), pattern(T,P,J).
pattern(_,_,_).

% All the generated premises can be proven.
% For all alternative bindings of Cond, Prem is a premise.
all(Cond, Prem) :-
  findall((Cond,Prem), Cond, Bag), all(Cond,Prem,Bag,Prems), premises(Prems).
all(Cond, Prem, [(Cond1,Prem1)|Bag], [Prem1|Prems]) :-
  unify_vars(Cond,Cond1), unify_vars(Prem,Prem1), all(Cond,Prem,Bag,Prems).
all(_, _, [], []).

unify_vars(Term1, Term2) :- var(Term1), var(Term2), !, Term1=Term2.
unify_vars(Term1, Term2) :- ground(Term1), ground(Term2), !, Term1=Term2.
unify_vars(Term1, Term2) :- (var(Term1) ; var(Term2)), !.
unify_vars(Term1, Term2) :- compound(Term1), compound(Term2),
  functor(Term1,F,N), functor(Term2,F,N), unify_vars(Term1,Term2,1,N).
unify_vars(_, _, I, N) :- succ(N,I).
unify_vars(Term1, Term2, I, N) :- arg(I,Term1,A1), arg(I,Term2,A2),
  unify_vars(A1,A2), succ(I,J), unify_vars(Term1,Term2,J,N).

% P is a conjunct in C.
% conjunct(?P, ?C)
conjunct(P, and(P,_)).
conjunct(P, C)        :- compound(C), C=and(_,B), var(B), !, P=B.
conjunct(P, and(_,C)) :- conjunct(P, C).
conjunct(P, P)        :- \+ conjunction(P).

conjunction(Term) :- compound(Term), Term=and(_,_).

% A is a closed subformula of B.
% csf(?A, +B)
csf(A, B) :- nonvar(B), A=B, \+ number(A).
csf(A, B) :- compound(B), arg(I,B,C), csf(A,C),
             forall((subterm(X,A), var(X)),
                    \+ (arg(J,B,D), I\=J, subterm(Y,D), X==Y)).
subterm(A, A).
subterm(A, B) :- compound(B), arg(_,B,C), subterm(A,C).

% ARITHMETIC

% Solve equation A=X where A is an integer and
%                          X is an expression with one variable.
% eq(+A, +X)
eq(A,X) :- A=X, !.
eq(A,Y) :- (Y=X+B ; Y=B+X), integer(B), C is A-B, eq(C,X), !.
eq(A,X-B) :- integer(B), C is A+B, eq(C,X), !.
eq(A,B-X) :- integer(B), C is B-A, eq(C,X), !.
eq(A,-X) :- B is -A, eq(B,X), !.

% Evalute all expressions in In.
% eval(+In, -Out)
eval(Expr,Number) :- operation(Expr), expression(Expr), Number is Expr, !.
eval(In,Out) :- operation(In), term_variables(In,[Out]), copy_term(In,X),
                term_variables(X,[0]), eq(0,X), !.
eval(In,Out) :-
  compound(In), arg(I,In,A1), eval(A1,A2), A1 \== A2, !,
  functor(In,Name,N), functor(Out,Name,N), arg(I,Out,A2), eval(In,Out,1,I), !.
eval(Term,Term).
eval(In,Out,I,I) :- succ(I,J), eval(In,Out,J,0).
eval(In,Out,I,X) :- arg(I,In,A1), !, eval(A1,A2), arg(I,Out,A2),
                    succ(I,J), eval(In,Out,J,X).
eval(_,_,_,0).

% Use eq/2 to bind free variables, evaluate expressions and compare or unify
% the two terms, recursively.
% equals(+Term1, +Term2)
equals(T1,T2) :- T1=T2, !.
equals(T1,T2) :- integer(T1), expression(T2), !, T1=:=T2.
equals(T1,T2) :- integer(T1), eval(T2,X), eq(T1,X), !.
equals(T1,T2) :- expression(T2), Number is T2, equals(Number,T1), !.
equals(T1,T2) :- compound(T1), compound(T2), functor(T1,F,N), functor(T2,F,N),
                 arg(I,T1,A1), arg(I,T2,A2), A1\==A2, equals(A1,A2),
                 equals(T1,T2,1,I), !.
equals(T1,T2,I,I) :- !, succ(I,J), equals(T1,T2,J,0).
equals(T1,T2,I,X) :- arg(I,T1,A1), !, arg(I,T2,A2), equals(A1,A2),
                     succ(I,J), equals(T1,T2,J,X).
equals(_,_,_,0).

% True if Term is an integer or arithmetic function of expressions.
% expression(+Term)
expression(Var) :- var(Var), !, fail.
expression(Int) :- integer(Int).
expression(-Expr) :- expression(Expr).
expression(Expr) :- operation(Expr,A,B), expression(A), expression(B).

operation(Term) :- \+ compound(Term), !, fail.
operation(-_).
operation(Term) :- operation(Term,_,_).
operation(A+B, A,B).
operation(A-B, A,B).
operation(max(A,B), A,B).
operation(min(A,B), A,B).

% DATABASE
% These predicates use the database as a set without duplicate facts.
% The hash table (bucketarray) is used to determine if a fact is in the set.

% Assert Fact in the database and add it to the hash table.
% Fails if Fact is already in the hash table.
new_fact(Fact) :- hash(Fact,Hash), lookup(Hash,Bucket),
                  \+ element_of(Fact, Bucket),
                  record_bucket(Hash, [Fact|Bucket]),
                  assertz(Fact).

% True if Fact is an element of the database set.
element(Fact) :- hash(Fact,Hash), lookup(Hash,Bucket),
                 element_of(Fact, Bucket).

% Delete all facts that unifies with Term.
delete_facts(Term) :- copy_term(Term,Fact),
  (retract(Fact), hash(Fact,Hash), lookup(Hash,Bucket),
   delete_entries(Hash,Bucket,Term), fail ; true).
delete_entries(_,[],_) :- !.
delete_entries(Hash,Bucket,Term) :- delete(Bucket, Term, NewBucket),
                                    record_bucket(Hash, NewBucket).

hash(Term, Hash) :- term_hash(Term,Hash), integer(Hash), !.
hash(Term, Hash) :- copy_term(Term,T), numbervars(T,0,_), term_hash(T,Hash).

% Check whether Element is in Set without unifying any variables,
% element_of(+Element, +Set)
element_of(_, []) :- !, fail.
element_of(X, Set) :- ground(X), !, member(Y, Set), X==Y, !.
element_of(X, Set) :- copy_term(X,Y), member(Y, Set), copy_term(X,Z),
                      term_variables(Y, Vars),
                      term_variables(Z, Vars), Y==Z, !.

% PRINTING

print_numvars(Term) :- current_output(Out), print_numvars(Out,Term).
print_numvars(Stream, Term1) :-
  copy_term(Term1, Term2), numbervars(Term2,23,_,[singletons]), print(Stream, Term2).

% portray belief
portray(B) :- current_predicate(time_step/2),
  time_step(B,I), arg(TimeArg,B,V), V==I, print(I), write(' :'), 
  forall((arg(Arg,B,X),Arg\=TimeArg),
         (write(' '), print_formula(X))). 

print_formula(X) :- var(X), !, print(X).
print_formula(A->B) :- !, write('('), print_formula(A), write(' -> '),
                                      print_formula(B), write(')').
print_formula(and(A,C)) :-
  write('('), print_formula(A), conjunction(C), !,
  forall(conjunct(P,C), (write(' and '), print_formula(P))),
  write(')').
print_formula(and(_,B)) :- !, write(' and '), print_formula(B), write(')').
print_formula(not(P)) :- !, write('not '), print_formula(P).
print_formula(P) :- print(P).

% vim: syntax=prolog sts=2 sw=2 expandtab:
