#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include "input.h"
#include "timer.h"
#include "tetris.h"
#include "textgfx/textgfx.h"
#include "prolog-io.h"

static struct termios saved_term_attr;

/* Non-canonical terminal input mode */
void set_input_mode()
{
	struct termios attr;
	tcgetattr(STDIN_FILENO, &saved_term_attr);
	tcgetattr(STDIN_FILENO, &attr);
	attr.c_lflag &= ~(ICANON|ECHO);
	attr.c_cc[VMIN] = 0;
	attr.c_cc[VTIME] = 2;
	tcsetattr(STDIN_FILENO, TCSADRAIN, &attr);
}

void restore_input_mode()
{
	tcsetattr(STDIN_FILENO, TCSANOW, &saved_term_attr);
}

static int select_read(int nfds, fd_set *readfds, int ms)
{
	struct timeval tmv;
	int ret;
	tmv.tv_sec = ms / 1000;
	tmv.tv_usec = 1000 * (ms % 1000);
	ret = select(nfds, readfds, NULL, NULL, &tmv);
	if (ret < 0 && errno != EINTR) {
		textgfx_end();
		perror("select");
		exit(1);
	}
	return ret;
}

static int select1(int fd, int ms)
{
	fd_set set;
	int ret;
	if (ms < 0)
		return 0;
	FD_ZERO(&set);
	FD_SET(fd, &set);
	ret = select_read(fd+1, &set, ms);
	if (ret >= 0)
		return ret;
	return select1(fd, ms-1);
}

int readchar(int fd, char *c)
{
	if (select1(fd, 0))
		return read(fd, c, 1);
	return 0;
}

static int waitforinput(fd_set *set, int ms)
{
	int ret;
	if (ms < 0)
		return 0;
	FD_ZERO(set);
	FD_SET(STDIN_FILENO, set);
	FD_SET(prolog_out_r, set);
	ret = select_read(prolog_out_r+1, set, ms);
	if (ret >= 0)
		return ret;
	return waitforinput(set, ms-1);
}

static int keytoaction(int key, int flags)
{
	struct player *plr = &player1;
	int rot    = 0;
	int rot_cw = 0;
	key &= 0x7f;
	switch (key) {
	case '\n':
		return 1;
	case ESC:
	case '\b':
		game.state = GAME_CREATED;
		return -2;
	case A_BTN:
		rot_cw = 1;
	case B_BTN:
		rot = 1;
	}
	if (flags & DISCARD_MOVES) {
		if (rot)
			initialrotate(plr, rot_cw);
		return -1;
	}
	if (!(flags & DISCARD_DROPS)) {
		switch (key) {
		case HARDDROP:
			return harddrop(plr, 0);
		case MVDOWN:
			if (!movedown(plr, 1))
				return 0;
			return 1;
		}
	}
	switch (key) {
	case MVLEFT:
		moveleft(plr);
		break;
	case MVRIGHT:
		moveright(plr);
		break;
	default:
		if (rot) {
			rotate(plr, rot_cw);
			break;
		}
		return -1;
	}
	return 1;
}

int do_action(int tm, int flags)
{
	fd_set set;
	int t;
	int key;
	while (tm > 0) {
		t = gettm(0);
		key = 0;
		editcommandline();
		if (waitforinput(&set, tm)) {
			if (FD_ISSET(prolog_out_r, &set))
				key = read_prolog_output();
		}
		if (key)
			return keytoaction(key, flags);
		if (gettm(t) == t)
			sleep_msec(1);
		tm -= gettm(t)-t;
	}
	return -1;
}

int waitforstart()
{
	int ret = 0;
	while (ret != 1) {
		ret = do_action(1000, DISCARD_MOVES);
		if (ret == -2)
			return 0;
	}
	return 1;
}
