#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "prolog-io.h"
#include "input.h"
#include "textgfx/textgfx.h"

int prolog_in_w;	/* write end of stdin pipe */
int prolog_out_r;	/* read end of stdout pipe */

int prolog_window_margin = 34;

#define WINDOW_H 21

static char prologwindow[WINDOW_H][80];
static int  curx, cury;

static char commandline[80];

static pid_t prolog_pid;

int spawn_prolog(int argc, char **argv)
{
	char pl[] = "pl";
	int ipipefd[2];
	int opipefd[2];
	pid_t pid;
	if (pipe(ipipefd) < 0 || pipe(opipefd) < 0) {
		perror("pipe");
		return 0;
	}
	pid = fork();
	if (pid < 0) {
		perror("fork");
		return 0;
	}
	if (pid == 0) {
		close(ipipefd[1]);
		close(opipefd[0]);
		dup2(ipipefd[0], 0);
		dup2(opipefd[1], 1);
		dup2(opipefd[1], 2);
		if (argc > 1)
			argv++;
		else
			argv[0] = pl;
		int i;
		for (i=0; argv[i]; i++) {
			printf("%s\n", argv[i]);
		}
		execvp(argv[0], argv);
		return 0;
	} 
	prolog_pid = pid;
	close(ipipefd[0]);
	close(opipefd[1]);
	prolog_in_w  = ipipefd[1];
	prolog_out_r = opipefd[0];
	return 1;
}

void stop_prolog()
{
	int status;
	close(prolog_in_w);
	close(prolog_out_r);
	waitpid(prolog_pid, &status, 0);
}

static void beginline(int y)
{
	setwin(0);
	setattr_normal();
	setcurs(prolog_window_margin, y);
	cleartoeol();
}

static void printline(int i)
{
	beginline(i);
	printstr(prologwindow[i]);
}

static void scroll()
{
	int i;
	for (i=0; i<WINDOW_H-1; i++) {
		memcpy(prologwindow[i], prologwindow[i+1], 80);
		printline(i);
	}
	prologwindow[i][0] = '\0';
	printline(i);
	cury--;
}

static int scrollto(int i)
{
	for (; i >= WINDOW_H; i--)
		scroll();
	return i;
}

static void put(int c)
{
	int w = term_width - prolog_window_margin;
	int i = cury;
	int j = curx;

	if (w > 80)
		w = 80;
	if (c == '\t')
		c = ' ';
	if (c == '\n' && i > 0 && !prologwindow[i-1][0] && !prologwindow[i][0])
		return;
	if (c == '\n' || j >= w-1) {
		if (i < WINDOW_H)
			printline(i);
		i++; j=0;
	}
	if (isprintable(c)) {
		i = scrollto(i);
		prologwindow[i][j] = c;
		j++;
		prologwindow[i][j] = '\0';
		printline(i);
	}
	cury = i;
	curx = j;
}

static int startswith(const char *s, const char *prefix)
{
	return *prefix =='\0' || *s == *prefix && startswith(s+1, prefix+1); 
}

static int linestartswith(int i, const char *prefix)
{
	return i>=0 && startswith(prologwindow[i], prefix);
}

static void reformat_action(const char *s)
{
	char *line;
	if (cury <= 0)
		return;
	cury--;
	line = prologwindow[cury];
	curx = strlen(line);
	put(' ');
	for (; *s; s++)
		put(*s);
}

int read_prolog_output()
{
	static const struct {
		const char *from;
		int         to;
	} map[] = {
		{":start", '\n'},
		{":quit", ESC},
		{":left", MVLEFT},
		{":right", MVRIGHT},
		{":down", MVDOWN},
		{":rotate", A_BTN},
		{":\\rotate", B_BTN},
		{":drop", HARDDROP},
		{NULL}
	};
	static int started;
	char *line;
	char c;
	int i;
	while (readchar(prolog_out_r, &c)) {
		put(c);
		line = prologwindow[cury];
		if (line[0] == ':') {
			for (i=0; map[i].from; i++)
				if (!strcmp(line, map[i].from)) {
					line[0] = '\0';
					reformat_action(map[i].from);
					started = 1;
					return map[i].to;
				}
		}
		if (!strcmp(line, "true.")) {
			line[0] = '\0';
			curx = 0;
		}
	}
	for (i=0; i<cury; i++)
		if (linestartswith(i, "ERROR:") && started)
			return ESC;
	return 0;
}

static void write_prolog(const char *s, int n)
{
	while (write(prolog_in_w, s, n) < 0 && errno==EINTR)
		;
}

void write_prolog_str(const char *s)
{
	write_prolog(s, strlen(s));
}

void prolog_command(const char *line)
{
	int cut = 0;
	int n = strlen(line);
	int i;
	while (n > 0 && line[n-1] == ' ')
		n--;
	if (n <= 0)
		return;
	write_prolog(line, n);
	if (line[n-1] == '.')
		n--;
	else {
		write_prolog(",!.", 3);
		cut = 1;
	}
	write_prolog("\n", 1);
	if (curx > 0)
		put('\n');
	put('?'); put('-'); put(' ');
	for (i=0; i<n; i++)
		put(line[i]);
	if (cut) {
		put(',');
		put('!');
	}
	put('.'); put('\n');
}

void editcommandline()
{
	int n = strlen(commandline);
	char c;
	if (readchar(0, &c)) {
		if (isprintable(c) && n+1 < sizeof(commandline)) {
			commandline[n] = c;
			n++;
		}
		if ((c == DEL || c == '\b') && n > 0)
			n--;
		if (c == '\n') {
			prolog_command(commandline);
			n = 0;
		}
		commandline[n] = '\0';
	}
	beginline(scrollto(cury));
	printstr("?- ");
	printstr(commandline);
	fflush(stdout);
}
